from pathlib import Path

from brainglobe_utils.IO import yaml

data_dir = Path("tests", "data")
yaml_file = data_dir / "yaml" / "single_section.yml"

yaml_section = [
    {"a": "/some_path", "b": "Eeyore", "d": "a_single_dog", "e": 2},
    {"a": "/another/path", "z": 1, 2: 2},
]

yaml_contents = {"data": yaml_section}


def test_open_yaml():
    yaml_contents_test = yaml.open_yaml(yaml_file)
    assert yaml_contents_test == yaml_contents


def test_read_yaml_section(section="data"):
    yaml_section_test = yaml.read_yaml_section(yaml_file, section)
    assert yaml_section_test == yaml_section


def test_save_yaml(tmpdir):
    test_yaml_path = Path(tmpdir) / "test.yml"
    yaml.save_yaml(yaml_contents, test_yaml_path)
    test_yaml = yaml.open_yaml(test_yaml_path)
    saved_yaml = yaml.open_yaml(yaml_file)
    assert test_yaml == saved_yaml



from pathlib import Path

from brainglobe_utils.general.pathlib import append_to_pathlib_stem


def test_append_to_pathlib_stem():
    path = Path("path", "to", "file.txt")
    appended_path = append_to_pathlib_stem(path, "_appended")
    assert appended_path == Path("path", "to", "file_appended.txt")



import numpy as np

import brainglobe_utils.image.masking as masking

raw_image = np.array(
    [
        [1, 1, 3, 3, 1, 1],
        [1, 1, 5, 5, 1, 1],
        [1, 1, 5, 5, 1, 1],
        [1, 1, 5, 5, 1, 1],
    ]
)

mask_val_4 = np.array(
    [
        [0, 0, 0, 0, 0, 0],
        [0, 0, 1, 1, 0, 0],
        [0, 0, 1, 1, 0, 0],
        [0, 0, 1, 1, 0, 0],
    ]
)

masked_image = np.array(
    [
        [0, 0, 0, 0, 0, 0],
        [0, 0, 5, 5, 0, 0],
        [0, 0, 5, 5, 0, 0],
        [0, 0, 5, 5, 0, 0],
    ]
)


def test_make_mask():
    assert (mask_val_4 == masking.make_mask(raw_image, threshold=4)).all()


def test_mask_image_threshold():
    result = masking.mask_image_threshold(raw_image, raw_image, threshold=4)
    assert (result == masked_image).all()



import numpy as np

from brainglobe_utils.image import scale

test_2d_img = np.array([[1, 2, 10, 100], [5, 25, 300, 1000], [1, 0, 0, 125]])
validate_2d_img = np.array(
    [
        [65.535, 131.07, 655.35, 6553.5],
        [327.675, 1638.375, 19660.5, 65535],
        [65.535, 0, 0, 8191.875],
    ]
)


def test_scale_to_16_bits():
    validate_2d_img_uint16 = validate_2d_img.astype(np.uint16, copy=False)
    assert (
        validate_2d_img_uint16
        == scale.scale_and_convert_to_16_bits(test_2d_img)
    ).all()



import yaml


def read_yaml_section(yaml_file, section):
    yaml_contents = open_yaml(yaml_file)
    contents = yaml_contents[section]
    return contents


def open_yaml(yaml_file):
    with open(yaml_file) as f:
        yaml_contents = yaml.load(f, Loader=yaml.FullLoader)
    return yaml_contents


def save_yaml(yaml_contents, output_file, default_flow_style=False):
    with open(output_file, "w") as outfile:
        yaml.dump(
            yaml_contents, outfile, default_flow_style=default_flow_style
        )



import numpy as np


def mask_image_threshold(image, masking_image, threshold=0):
    """
    Mask one image, based on the values in another image that are above a
    threshold
    :param image: Input image
    :param masking_image: Image to base the mask on (same shape as image)
    :param threshold: Threshold to base the mask on
    :return: Masked image
    """
    masking_image = make_mask(masking_image, threshold=threshold)
    return image * masking_image


def make_mask(masking_image, threshold=0):
    """
    Given an image, and an optional threshold, returns a binary image that can
    be used as a mask
    :param masking_image: nD image
    :param threshold: Optional threshold, default 0. Values above this are
    included in the mask. Values below this are not.
    :return: Binary mask
    """
    mask = np.copy(masking_image)
    mask[masking_image <= threshold] = 0
    mask[masking_image > threshold] = 1
    return mask



import os

from natsort import natsorted

from brainglobe_utils.cells import cells
from brainglobe_utils.IO.cells import get_cells

data_dir = os.path.join("tests", "data")
cubes_dir = os.path.join(data_dir, "cube_extract", "cubes")
xml_path = os.path.join(data_dir, "cells", "cells.xml")


def test_pos_from_file_name():
    positions_validate = [
        [392, 522, 10],
        [340, 1004, 15],
        [340, 1004, 15],
        [392, 522, 10],
    ]
    cube_files = os.listdir(cubes_dir)
    positions = []
    for file in cube_files:
        positions.append(cells.pos_from_file_name(file))
    assert natsorted(positions) == natsorted(positions_validate)


def test_group_cells_by_z():
    z_planes_validate = [
        1272,
        1273,
        1274,
        1275,
        1276,
        1277,
        1278,
        1279,
        1280,
        1281,
        1282,
        1283,
        1284,
        1285,
        1286,
        1287,
        1288,
        1289,
        1290,
        1291,
        1292,
        1294,
        1295,
        1296,
        1297,
        1298,
    ]

    cell_numbers_in_groups_validate = [
        1,
        3,
        7,
        8,
        3,
        1,
        4,
        3,
        1,
        2,
        2,
        1,
        1,
        2,
        5,
        2,
        2,
        2,
        3,
        1,
        1,
        6,
        1,
        1,
        1,
        1,
    ]

    cell_list = get_cells(xml_path)
    cells_groups = cells.group_cells_by_z(cell_list)
    z_planes_test = list(cells_groups.keys())
    z_planes_test.sort()

    assert z_planes_validate == z_planes_test

    cell_numbers_in_groups_test = [
        len(cells_groups[plane]) for plane in z_planes_test
    ]
    assert cell_numbers_in_groups_validate == cell_numbers_in_groups_test



class CommandLineInputError(Exception):
    """Exception raised for incorrect or illogical command line inputs that
    are not caught elsewhere.

    Attributes:
        expression -- input expression in which the error occurred
        message -- explanation of the error
    """

    def __init__(self, message):
        self.message = message

    def __str__(self):
        return str(self.message)



import filecmp
from pathlib import Path

import pytest

from brainglobe_utils.brainmapper.analysis import (
    Point,
    count_points_per_brain_region,
)


@pytest.fixture
def points():
    p1 = Point(
        raw_coordinate=[1.0, 1.0, 1.0],
        atlas_coordinate=[1.0, 1.0, 1.0],
        structure="Paraventricular hypothalamic nucleus, descending division",
        structure_id=56,
        hemisphere="right",
    )
    p2 = Point(
        raw_coordinate=[2.0, 2.0, 2.0],
        atlas_coordinate=[2.0, 2.0, 2.0],
        structure="Anterodorsal nucleus",
        structure_id=57,
        hemisphere="left",
    )
    return [p1, p2]


@pytest.fixture
def structures_with_points():
    return [
        "Paraventricular hypothalamic nucleus, descending division",
        "Anterodorsal nucleus",
    ]


def test_get_region_totals(tmp_path, points, structures_with_points):
    """Regression test for count_points_per_brain_region for
    pandas 1.5.3 -> 2.1.3+.
    pd.Dataframe.append was deprecated and removed in this time."""
    OUTPUT_DATA_LOC = (
        Path(__file__).parent.parent.parent / "data" / "brainmapper"
    )

    volumes_path = OUTPUT_DATA_LOC / "volumes.csv"
    expected_output = (
        OUTPUT_DATA_LOC / "region_totals_regression_pandas1_5_3.csv"
    )

    output_path = Path(tmp_path / "tmp_region_totals.csv")
    count_points_per_brain_region(
        points, structures_with_points, volumes_path, output_path
    )
    assert output_path.exists()
    assert filecmp.cmp(output_path, expected_output)



from pathlib import Path
from typing import Tuple, Union

import imio
import numpy as np
from scipy.ndimage import zoom
from skimage.filters import gaussian

from brainglobe_utils.general.system import ensure_directory_exists
from brainglobe_utils.image.binning import get_bins
from brainglobe_utils.image.masking import mask_image_threshold
from brainglobe_utils.image.scale import scale_and_convert_to_16_bits


def rescale_array(source_array, target_array, order=1):
    """
    Rescale a source array to the size of a target array.

    Parameters
    ----------
    source_array : np.ndarray
        The array to be rescaled.
    target_array : np.ndarray
        The array whose size will be matched.
    order : int (default=1)
    Returns
    -------
    np.ndarray
        The rescaled source array.
    """

    # Compute the zoom factors for each dimension
    zoom_factors = [
        t / s for s, t in zip(source_array.shape, target_array.shape)
    ]

    # Use scipy's zoom function to rescale the array
    rescaled_array = zoom(source_array, zoom_factors, order=order)

    return rescaled_array


def heatmap_from_points(
    points: np.ndarray,
    image_resolution: float,
    image_shape: Tuple[int, int, int],
    bin_sizes: Tuple[int, int, int] = (1, 1, 1),
    output_filename: Union[str, Path, None] = None,
    smoothing: Union[float, None] = None,
    mask_image: Union[np.ndarray, None] = None,
) -> np.ndarray:
    """
    Generate a heatmap from a set of points and a reference image that
    defines the shape and resolution of the heatmap.

    Parameters
    ----------
    points : np.ndarray
        Data points to be used for heatmap generation.
        These may be cell positions, e.g. from cellfinder.
    image_resolution : float
        Resolution of the image to be generated.
    image_shape : Tuple[int, int, int]
        The shape of the downsampled heatmap as a tuple of integers
    bin_sizes : Tuple[int, int, int], optional
        The bin sizes for each dimension of the heatmap (default is (1, 1, 1)).
    output_filename : Union[str, pathlib.Path, None]
        The filename or path where the heatmap image will be saved.
        Can be a string or a pathlib.Path object. If None, the heatmap will
        not be saved
    smoothing : float, optional
        Smoothing factor to be applied to the heatmap. Expressed in "real"
        units (e.g. microns). This will be converted to voxel units based on
        the image_resolution.
        If None, no smoothing is applied (default is None).
    mask_image : Union[np.ndarray, None], optional
        An optional mask image (as a NumPy array) to apply to the heatmap.
        This could be e.g. a registered atlas image (so points outside the
        brain are masked). The image will be
        binarised, so this image can be binary or not.
         If None, no mask is applied (default is None).

    Returns
    -------
    np.ndarray
        The generated heatmap as a NumPy array.
    """

    bins = get_bins(image_shape, bin_sizes)
    heatmap_array, _ = np.histogramdd(points, bins=bins)
    heatmap_array = heatmap_array.astype(np.uint16)

    if smoothing is not None:
        smoothing = int(round(smoothing / image_resolution))
        heatmap_array = gaussian(heatmap_array, sigma=smoothing)

    if mask_image is not None:
        if mask_image.shape != heatmap_array.shape:
            mask_image = rescale_array(mask_image, heatmap_array)

        heatmap_array = mask_image_threshold(heatmap_array, mask_image)

    heatmap_array = scale_and_convert_to_16_bits(heatmap_array)

    if output_filename is not None:
        ensure_directory_exists(Path(output_filename).parent)
        imio.to_tiff(heatmap_array, output_filename)

    return heatmap_array



from importlib.metadata import PackageNotFoundError, version

__author__ = "Adam Tyson"
try:
    __version__ = version("brainglobe-utils")
except PackageNotFoundError:
    # package is not installed
    pass



import numpy as np

from brainglobe_utils.image import binning


def test_get_bins():
    image_size = (1000, 20, 125, 725)
    bin_sizes = (500, 2, 25, 100)

    dim0_bins = np.array((0, 500, 1000))
    dim1_bins = np.array((0, 2, 4, 6, 8, 10, 12, 14, 16, 18, 20))
    dim2_bins = np.array((0, 25, 50, 75, 100, 125))
    dim3_bins = np.array((0, 100, 200, 300, 400, 500, 600, 700))
    bins = binning.get_bins(image_size, bin_sizes)

    assert (dim0_bins == bins[0]).all()
    assert (dim1_bins == bins[1]).all()
    assert (dim2_bins == bins[2]).all()
    assert (dim3_bins == bins[3]).all()



from pathlib import Path

import numpy as np
import pytest
from tifffile import imread

from brainglobe_utils.image.heatmap import heatmap_from_points, rescale_array

data_dir = Path("tests", "data")
heatmap_validate_path = data_dir / "image" / "heatmap.tif"

points = np.array([[5, 5, 5], [10, 10, 10], [15, 15, 15]])


@pytest.fixture
def mask_array():
    """
    A simple array of size (20,20,20) where the left side
    is 0 and the right side is 1
    """
    array = np.zeros((20, 20, 20))
    array[:, :, 10:] = 1
    return array


def test_rescale_array(mask_array):
    small_array = np.zeros((10, 10, 10))
    resized_array = rescale_array(mask_array, small_array)
    assert resized_array.shape == small_array.shape


def test_heatmap_from_points(tmp_path, mask_array):
    output_filename = tmp_path / "test_heatmap.tif"

    heatmap_test = heatmap_from_points(
        points,
        2,
        (20, 20, 20),
        bin_sizes=(2, 2, 2),
        output_filename=output_filename,
        smoothing=5,
        mask_image=mask_array,
    )
    heatmap_validate = imread(heatmap_validate_path)
    assert np.array_equal(heatmap_validate, heatmap_test)

    heatmap_file = imread(output_filename)
    assert np.array_equal(heatmap_validate, heatmap_file)



def marching_cubes_to_obj(marching_cubes_out, output_file):
    """
    Saves the output of skimage.measure.marching_cubes as an .obj file

    :param marching_cubes_out: tuple
    :param output_file: str

    """

    verts, faces, normals, _ = marching_cubes_out
    with open(output_file, "w") as f:
        for item in verts:
            f.write(f"v {item[0]} {item[1]} {item[2]}\n")
        for item in normals:
            f.write(f"vn {item[0]} {item[1]} {item[2]}\n")
        for item in faces:
            f.write(
                f"f {item[0]}//{item[0]} {item[1]}//{item[1]} "
                f"{item[2]}//{item[2]}\n"
            )
        f.close()



import logging


def suppress_specific_logs(logger, message):
    logger = logging.getLogger(logger)

    class NoParsingFilter(logging.Filter):
        def filter(self, record):
            return not record.getMessage().startswith(message)

    logger.addFilter(NoParsingFilter())



import numpy as np
import pandas as pd

from brainglobe_utils.pandas import misc as pandas_misc

columns = ["name", "number"]
data_with_nan = [["one", np.nan], ["two", 15], ["three", np.nan]]
df_with_nan = pd.DataFrame(data_with_nan, columns=columns)
data_with_inf = [["one", np.inf], ["two", 15], ["three", np.inf]]
df_with_inf = pd.DataFrame(data_with_inf, columns=columns)


def test_initialise_df():
    test_df = pandas_misc.initialise_df("one", "two", "3")
    df = pd.DataFrame(columns=["one", "two", "3"])
    assert df.equals(test_df)


def test_sanitise_df():
    sanitised_df = pandas_misc.sanitise_df(df_with_inf)
    assert sanitised_df.equals(df_with_nan)


def test_safe_pandas_concat() -> None:
    """
    Test the following:
    - Non-empty dataframes are concatenated as expected,
    - When one dataframe is empty, the other is returned,
    - When both dataframes are empty, an empty dataframe with
    the corresponding columns is returned.
    """
    df1 = pd.DataFrame(data={"a": [1], "b": [2], "c": [3]})
    df2 = pd.DataFrame(data={"a": [4], "b": [5], "c": [6]})
    empty_df = pd.DataFrame(columns=["a", "b", "c"])
    combined_df = pd.DataFrame(data={"a": [1, 4], "b": [2, 5], "c": [3, 6]})

    assert combined_df.equals(pandas_misc.safe_pandas_concat(df1, df2))
    assert df1.equals(pandas_misc.safe_pandas_concat(df1, empty_df))
    assert df2.equals(pandas_misc.safe_pandas_concat(empty_df, df2))
    assert empty_df.equals(pandas_misc.safe_pandas_concat(empty_df, empty_df))



import os

import pandas as pd
import pytest
from natsort import natsorted

from brainglobe_utils.cells.cells import Cell
from brainglobe_utils.IO import cells as cell_io

data_dir = os.path.join("tests", "data")
xml_path = os.path.join(data_dir, "cells", "cells.xml")
yml_path = os.path.join(data_dir, "cells", "cells.yml")
cubes_dir = os.path.join(data_dir, "cube_extract", "cubes")
roi_sorter_output_dir = os.path.join(data_dir, "IO", "roi_sorter_output")

type_vals = [
    1,
    1,
    1,
    1,
    1,
    1,
    1,
    1,
    1,
    1,
    1,
    1,
    1,
    1,
    1,
    1,
    1,
    1,
    1,
    1,
    1,
    1,
    1,
    1,
    1,
    1,
    1,
    1,
    1,
    1,
    1,
    1,
    1,
    1,
    1,
    1,
    1,
    1,
    1,
    1,
    1,
    1,
    1,
    1,
    1,
    1,
    1,
    1,
    1,
    1,
    1,
    1,
    1,
    1,
    1,
    1,
    1,
    1,
    1,
    1,
    1,
    1,
    1,
    1,
    1,
]
x_vals = [
    181,
    2117,
    4082,
    2137,
    2129,
    2141,
    2130,
    2130,
    2133,
    2141,
    2162,
    2127,
    1801,
    2176,
    2194,
    2664,
    2435,
    1266,
    568,
    2103,
    2098,
    2105,
    2208,
    2162,
    2221,
    2081,
    2230,
    2015,
    2072,
    1413,
    1950,
    3340,
    2051,
    3362,
    2046,
    2046,
    2046,
    2047,
    2032,
    2923,
    2064,
    2267,
    2297,
    2004,
    2356,
    2357,
    2359,
    2360,
    2365,
    2612,
    2623,
    2657,
    2906,
    2165,
    2685,
    2892,
    2896,
    2894,
    2901,
    2270,
    1924,
    2505,
    2531,
    2542,
    2536,
]
y_vals = [
    2379,
    133,
    332,
    134,
    133,
    134,
    132,
    133,
    136,
    136,
    130,
    1026,
    336,
    134,
    134,
    336,
    343,
    3427,
    2012,
    138,
    138,
    140,
    135,
    134,
    137,
    142,
    139,
    1045,
    141,
    2045,
    1048,
    142,
    142,
    145,
    150,
    151,
    151,
    151,
    146,
    147,
    147,
    147,
    1068,
    149,
    521,
    522,
    523,
    523,
    525,
    520,
    520,
    521,
    522,
    524,
    522,
    526,
    525,
    525,
    526,
    524,
    974,
    522,
    527,
    524,
    523,
]
z_vals = [
    1282,
    1276,
    1281,
    1286,
    1276,
    1278,
    1277,
    1278,
    1285,
    1297,
    1274,
    1294,
    1290,
    1279,
    1280,
    1296,
    1290,
    1294,
    1294,
    1285,
    1279,
    1289,
    1275,
    1286,
    1281,
    1284,
    1275,
    1291,
    1273,
    1294,
    1294,
    1274,
    1274,
    1274,
    1283,
    1287,
    1290,
    1294,
    1272,
    1275,
    1275,
    1274,
    1292,
    1273,
    1288,
    1286,
    1287,
    1289,
    1298,
    1274,
    1274,
    1273,
    1275,
    1275,
    1278,
    1286,
    1278,
    1288,
    1295,
    1279,
    1282,
    1275,
    1276,
    1275,
    1286,
]

cubes_cells = [
    Cell([340, 1004, 15], 1),
    Cell([340, 1004, 15], 1),
    Cell([392, 522, 10], 1),
    Cell([392, 522, 10], 1),
]

roi_sorter_cells = [
    Cell([4056, 564, 358], 1),
    Cell([3989, 267, 570], 1),
    Cell([4351, 735, 439], 1),
    Cell([4395, 677, 367], 1),
]


def test_get_cells():
    cells = cell_io.get_cells(xml_path)
    assert len(cells) == 65
    assert Cell([2536, 523, 1286], 1) == cells[64]

    cells = cell_io.get_cells(yml_path)
    assert len(cells) == 250
    assert Cell([9170, 2537, 311], 1) == cells[194]

    cells = cell_io.get_cells(cubes_dir)
    assert len(cells) == 4
    assert natsorted(cubes_cells) == natsorted(cells)

    cells = cell_io.get_cells(roi_sorter_output_dir)
    assert len(cells) == 4
    assert natsorted(roi_sorter_cells) == natsorted(cells)

    with pytest.raises(NotImplementedError):
        assert cell_io.get_cells("misc_format.abc")


def assert_cells_csv(csv_path):
    cells_df = pd.read_csv(csv_path)
    assert len(cells_df) == 65
    assert cells_df.type.tolist() == type_vals
    assert cells_df.x.tolist() == x_vals
    assert cells_df.y.tolist() == y_vals
    assert cells_df.z.tolist() == z_vals


def test_save_cells(tmp_path):
    cells = cell_io.get_cells(xml_path)
    tmp_cells_out_path = tmp_path / "cells.xml"
    cell_io.save_cells(cells, tmp_cells_out_path, save_csv=True)
    assert cells == cell_io.get_cells(str(tmp_cells_out_path))

    tmp_cells_out_path = tmp_path / "cells.csv"
    cell_io.cells_to_csv(cells, tmp_cells_out_path)
    assert_cells_csv(tmp_cells_out_path)


def test_cells_to_xml(tmp_path):
    cells = cell_io.get_cells(xml_path)
    tmp_cells_out_path = tmp_path / "cells.xml"
    cell_io.cells_to_xml(cells, tmp_cells_out_path)
    assert cells == cell_io.get_cells(str(tmp_cells_out_path))


def test_cells_xml_to_dataframe():
    cells_df = cell_io.cells_xml_to_df(xml_path)
    assert len(cells_df) == 65
    assert cells_df.type.tolist() == type_vals
    assert cells_df.x.tolist() == x_vals
    assert cells_df.y.tolist() == y_vals
    assert cells_df.z.tolist() == z_vals


def test_cells_to_csv(tmp_path):
    cells = cell_io.get_cells(xml_path)
    tmp_cells_out_path = tmp_path / "cells.csv"
    cell_io.cells_to_csv(cells, tmp_cells_out_path)
    assert_cells_csv(tmp_cells_out_path)



from pathlib import Path

from brainglobe_utils.general import string

data_dir = Path("tests", "data")
jabberwocky = data_dir / "general" / "jabberwocky.txt"

jabberwocky_list = [
    "’Twas brillig, and the slithy toves",
    "Did gyre and gimble in the wabe:",
    "All mimsy were the borogoves,",
    "And the mome raths outgrabe.",
    "“Beware the Jabberwock, my son!",
    "The jaws that bite, the claws that catch!",
    "Beware the Jubjub bird, and shun",
    "The frumious Bandersnatch!”",
    "He took his vorpal sword in hand;",
    "Long time the manxome foe he sought—",
    "So rested he by the Tumtum tree",
    "And stood awhile in thought.",
    "And, as in uffish thought he stood,",
    "The Jabberwock, with eyes of flame,",
    "Came whiffling through the tulgey wood,",
    "And burbled as it came!",
    "One, two! One, two! And through and through",
    "The vorpal blade went snicker-snack!",
    "He left it dead, and with its head",
    "He went galumphing back.",
    "“And hast thou slain the Jabberwock?",
    "Come to my arms, my beamish boy!",
    "O frabjous day! Callooh! Callay!”",
    "He chortled in his joy.",
    "’Twas brillig, and the slithy toves",
    "Did gyre and gimble in the wabe:",
    "All mimsy were the borogoves,",
    "And the mome raths outgrabe.",
]


def test_get_string_lines():
    assert string.get_text_lines(jabberwocky) == jabberwocky_list
    assert (
        string.get_text_lines(jabberwocky, return_lines=8)
        == jabberwocky_list[8]
    )



import numpy as np


def get_bins(image_size, bin_sizes):
    """
    Given an image size, and bin size, return a list of the bin boundaries
    :param image_size: Size of the final image (tuple/list)
    :param bin_sizes: Bin sizes corresponding to the dimensions of
    "image_size" (tuple/list)
    :return: List of arrays of bin boundaries
    """
    bins = []
    for dim in range(0, len(image_size)):
        bins.append(np.arange(0, image_size[dim] + 1, bin_sizes[dim]))
    return bins



import os
import platform
import random
from pathlib import Path
from random import shuffle
from unittest.mock import Mock, patch

import pytest

from brainglobe_utils.general import system
from brainglobe_utils.general.string import get_text_lines

data_dir = Path("tests", "data")
cubes_dir = data_dir / "cubes"
jabberwocky = data_dir / "general" / "jabberwocky.txt"
jabberwocky_sorted = data_dir / "general" / "jabberwocky_sorted.txt"

cubes = [
    "pCellz222y2805x9962Ch1.tif",
    "pCellz222y2805x9962Ch2.tif",
    "pCellz258y3892x10559Ch1.tif",
    "pCellz258y3892x10559Ch2.tif",
    "pCellz413y2308x9391Ch1.tif",
    "pCellz413y2308x9391Ch2.tif",
    "pCellz416y2503x5997Ch1.tif",
    "pCellz416y2503x5997Ch2.tif",
    "pCellz418y5457x9489Ch1.tif",
    "pCellz418y5457x9489Ch2.tif",
    "pCellz433y4425x7552Ch1.tif",
    "pCellz433y4425x7552Ch2.tif",
]

sorted_cubes_dir = [os.path.join(str(cubes_dir), cube) for cube in cubes]


@pytest.fixture
def mock_disk_usage():
    """Fixture to mock shutil.disk_usage."""
    return Mock(
        return_value=(1000000, 500000, 500000)
    )  # total, used, free space in bytes


@pytest.fixture
def mock_statvfs():
    """Fixture to mock os.statvfs."""
    mock_stats = Mock()
    mock_stats.f_frsize = 1024  # Fragment size
    mock_stats.f_bavail = 1000  # Free blocks
    return mock_stats


def test_replace_extension():
    test_file = "test_file.sh"
    test_ext = "txt"
    test_ext_w_dot = ".txt"
    validate_file = "test_file.txt"
    assert validate_file == system.replace_extension(test_file, test_ext)
    assert validate_file == system.replace_extension(test_file, test_ext_w_dot)


def test_remove_leading_character():
    assert ".ext" == system.remove_leading_character("..ext", ".")


def test_ensure_directory_exists(tmpdir):
    # string
    exist_dir = os.path.join(tmpdir, "test_dir")
    system.ensure_directory_exists(exist_dir)
    assert os.path.exists(exist_dir)
    os.rmdir(exist_dir)

    # pathlib
    exist_dir_pathlib = Path(tmpdir) / "test_dir2"
    system.ensure_directory_exists(exist_dir_pathlib)
    assert exist_dir_pathlib.exists()
    exist_dir_pathlib.rmdir()


def test_get_sorted_file_paths():
    # test list
    shuffled = sorted_cubes_dir.copy()
    shuffle(shuffled)
    assert system.get_sorted_file_paths(shuffled) == sorted_cubes_dir

    # test dir
    assert system.get_sorted_file_paths(cubes_dir) == sorted_cubes_dir
    assert (
        system.get_sorted_file_paths(cubes_dir, file_extension=".tif")
        == sorted_cubes_dir
    )

    # test text file
    # specifying utf8, as written on linux
    assert system.get_sorted_file_paths(
        jabberwocky, encoding="utf8"
    ) == get_text_lines(jabberwocky_sorted, encoding="utf8")

    # test unsupported
    with pytest.raises(NotImplementedError):
        system.get_sorted_file_paths(shuffled[0])


def test_check_path_in_dir():
    assert system.check_path_in_dir(jabberwocky, data_dir / "general")


def test_get_num_processes():
    cpu_count = 10
    with patch(
        "brainglobe_utils.general.system.psutil.cpu_count",
        return_value=cpu_count,
    ):
        assert system.get_num_processes(min_free_cpu_cores=0) == cpu_count


def test_max_processes():
    max_proc = 5
    correct_n = min(os.cpu_count(), max_proc)
    assert correct_n == system.get_num_processes(
        n_max_processes=max_proc, min_free_cpu_cores=0
    )


def test_max_processes_windows_low():
    cpu_count = 10
    with patch(
        "brainglobe_utils.general.system.platform.system",
        return_value="Windows",
    ):
        with patch(
            "brainglobe_utils.general.system.psutil.cpu_count",
            return_value=cpu_count,
        ):
            assert system.limit_cpus_windows(cpu_count) == cpu_count


def test_max_processes_windows_high():
    mock_cpu_count = 128
    with patch(
        "brainglobe_utils.general.system.platform.system",
        return_value="Windows",
    ):
        with patch(
            "brainglobe_utils.general.system.psutil.cpu_count",
            return_value=mock_cpu_count,
        ):
            # 61 is max on Windows
            assert system.limit_cpus_windows(mock_cpu_count) == 61


@pytest.mark.parametrize("cores_available", [1, 100, 1000])
def test_cores_available_in_slurm_environment(cores_available):
    mock_slurm_parameters = Mock()
    mock_slurm_parameters.allocated_cores = cores_available

    with patch.dict(
        "brainglobe_utils.general.system.os.environ", {"SLURM_JOB_ID": "1"}
    ), patch(
        "brainglobe_utils.general.system.slurmio.SlurmJobParameters",
        return_value=mock_slurm_parameters,
    ):
        assert system.get_cores_available() == cores_available


@pytest.mark.parametrize("cores_available", [1, 100, 1000])
def test_cores_available(cores_available):
    with patch(
        "brainglobe_utils.general.system.psutil.cpu_count",
        return_value=cores_available,
    ):
        assert system.get_cores_available() == cores_available


@pytest.mark.parametrize(
    "ram_needed_per_cpu, fraction_free_ram, max_ram_usage, "
    "free_system_ram, expected_cores",
    [
        (
            1024**3,
            0.1,
            None,
            16 * 1024**3,
            14,
        ),  # 1 GB per core, 0.1 fraction free, no max ram,
        # 16GB free on the system, expect 14
        (
            2 * 1024**3,
            0.5,
            None,
            256 * 1024**3,
            64,
        ),  # 1 GB per core, 0.5 fraction free, no max ram,
        # 256GB free on the system, expect 64
        (
            1024**3,
            0.5,
            10 * 1024**3,
            256 * 1024**3,
            5,
        ),  # 1 GB per core, 0.5 fraction free, 10Gb max ram,
        # 256GB free on the system, expect 5
    ],
)
def test_how_many_cores_with_sufficient_ram(
    ram_needed_per_cpu,
    fraction_free_ram,
    max_ram_usage,
    free_system_ram,
    expected_cores,
):
    with patch(
        "brainglobe_utils.general.system.get_free_ram",
        return_value=free_system_ram,
    ):
        assert (
            system.how_many_cores_with_sufficient_ram(
                ram_needed_per_cpu,
                fraction_free_ram,
                max_ram_usage=max_ram_usage,
            )
            == expected_cores
        )


def test_how_many_cores_with_sufficient_ram_in_slurm_environment():
    ram_needed_per_cpu = 1024**3  # 1 GB
    free_system_ram = 16 * 1024**3  # 16 GB

    mock_slurm_parameters = Mock()
    mock_slurm_parameters.allocated_memory = free_system_ram

    with patch.dict(
        "brainglobe_utils.general.system.os.environ", {"SLURM_JOB_ID": "1"}
    ), patch(
        "brainglobe_utils.general.system.slurmio.SlurmJobParameters",
        return_value=mock_slurm_parameters,
    ):
        assert (
            system.how_many_cores_with_sufficient_ram(ram_needed_per_cpu) == 14
        )  # (0.9 * 16) GB / 1 GB per core


def test_disk_free_gb_windows(mock_disk_usage):
    with patch(
        "brainglobe_utils.general.system.platform.system",
        return_value="Windows",
    ), patch(
        "brainglobe_utils.general.system.os.path.splitdrive",
        return_value=("C:\\", ""),
    ), patch(
        "brainglobe_utils.general.system.shutil.disk_usage", mock_disk_usage
    ):
        free_space = system.disk_free_gb("C:\\path\\to\\file")
        assert free_space == 500000 / 1024**3


@pytest.mark.skipif(
    platform.system() == "Windows",
    reason="os.statvfs not available on Windows",
)
def test_disk_free_gb_linux(mock_statvfs):
    with patch(
        "brainglobe_utils.general.system.platform.system", return_value="Linux"
    ), patch(
        "brainglobe_utils.general.system.os.statvfs", return_value=mock_statvfs
    ):
        free_space = system.disk_free_gb("/path/to/file")
        assert free_space == (1024 * 1000) / 1024**3  # Free space in GB


@pytest.mark.skipif(
    platform.system() == "Windows",
    reason="os.statvfs not available on Windows",
)
def test_disk_free_gb_macos(mock_statvfs):
    with patch(
        "brainglobe_utils.general.system.platform.system",
        return_value="Darwin",
    ), patch(
        "brainglobe_utils.general.system.os.statvfs", return_value=mock_statvfs
    ):
        free_space = system.disk_free_gb("/path/to/file")
        assert free_space == (1024 * 1000) / 1024**3  # Free space in GB


def test_get_free_ram():
    mock_free_ram = 1000000001

    mock_virtual_memory = Mock()
    mock_virtual_memory.available = mock_free_ram

    with patch(
        "brainglobe_utils.general.system.psutil.virtual_memory",
        return_value=mock_virtual_memory,
    ):
        assert system.get_free_ram() == mock_free_ram


def write_n_random_files(n, dir, min_size=32, max_size=2048):
    sizes = random.sample(range(min_size, max_size), n)
    for size in sizes:
        with open(os.path.join(dir, str(size)), "wb") as fout:
            fout.write(os.urandom(size))


def test_delete_directory_contents_with_progress(tmp_path):
    delete_dir = tmp_path / "delete_dir"
    os.mkdir(delete_dir)
    write_n_random_files(10, delete_dir)

    # check the directory isn't empty first
    assert not os.listdir(delete_dir) == []

    system.delete_directory_contents(delete_dir, progress=True)
    assert os.listdir(delete_dir) == []


def test_delete_directory_contents(tmp_path):
    delete_dir = tmp_path / "delete_dir"
    os.mkdir(delete_dir)
    write_n_random_files(10, delete_dir)

    # check the directory isn't empty first
    assert not os.listdir(delete_dir) == []

    system.delete_directory_contents(delete_dir, progress=False)
    assert os.listdir(delete_dir) == []



from pathlib import Path

from skimage import measure
from skimage.draw import ellipsoid

from brainglobe_utils.IO.surfaces import marching_cubes_to_obj

test_obj = (
    Path(__file__).parent.parent.parent / "data" / "IO" / "obj" / "test.obj"
)


def compare_text_files(a, b):
    with open(a, "r") as file1:
        a_contents = file1.readlines()
    with open(b, "r") as file2:
        b_contents = file2.readlines()
    assert a_contents == b_contents


def test_marching_cubes_to_obj(tmp_path):
    output_file = tmp_path / "test.obj"
    ellip_base = ellipsoid(6, 10, 16, levelset=True)
    marching_cubes_out = measure.marching_cubes(ellip_base, 0)
    marching_cubes_to_obj(marching_cubes_out, output_file)
    compare_text_files(output_file, test_obj)



def remove_empty_string(str_list):
    """
    Removes any empty strings from a list of strings
    :param str_list: List of strings
    :return: List of strings without the empty strings
    """
    return list(filter(None, str_list))


def unique_elements_lists(list_in):
    """return the unique elements in a list"""
    return list(dict.fromkeys(list_in))



"""
Cell position I/O

Based on https://github.com/SainsburyWellcomeCentre/niftynet_cell_count by
Christian Niedworok (https://github.com/cniedwor).
"""

import logging
import os
from typing import List, Optional
from xml.dom import minidom
from xml.etree import ElementTree
from xml.etree.ElementTree import Element as EtElement

import pandas as pd
import yaml

from brainglobe_utils.cells.cells import (
    Cell,
    MissingCellsError,
    UntypedCell,
    pos_from_file_name,
)
from brainglobe_utils.general.system import replace_extension


def get_cells(
    cells_file_path: str,
    cells_only: bool = False,
    cell_type: Optional[int] = None,
):
    if cells_file_path.endswith(".xml"):
        return get_cells_xml(cells_file_path, cells_only=cells_only)
    elif cells_file_path.endswith(".yml"):
        return get_cells_yml(cells_file_path, ignore_type=True)
    elif os.path.isdir(cells_file_path):
        try:
            return get_cells_dir(cells_file_path, cell_type=cell_type)
        except IndexError:
            # if a directory is given, but it contains
            # files that can't be read. Usually if the user gives the wrong
            # directory as input to `cellfinder_gen_cubes`
            raise_cell_read_error(cells_file_path)
    else:
        raise_cell_read_error(cells_file_path)


def raise_cell_read_error(cells_file_path):
    logging.error(
        "File format of: {} is not supported or contains errors. Please "
        "supply an xml file, or a directory of files with positions in the "
        "filenames."
        "".format(cells_file_path)
    )
    raise NotImplementedError(
        "File format of: {} is not supported or contains errors. Please "
        "supply an xml file, or a directory of files with positions in the "
        "filenames."
        "".format(cells_file_path)
    )


def get_cells_xml(xml_file_path, cells_only=False):
    with open(xml_file_path, "r") as xml_file:
        root = ElementTree.parse(xml_file).getroot()
        cells = []
        for type_marker in root.find("Marker_Data").findall("Marker_Type"):
            cell_type = int(type_marker.find("Type").text)
            for cell_element in type_marker.findall("Marker"):
                cells.append(Cell(cell_element, cell_type))
        if not cells:
            raise MissingCellsError(
                "No cells found in file {}".format(xml_file_path)
            )
    if cells_only:
        cells = [c for c in cells if c.is_cell()]
    return cells


def get_cells_yml(cells_file_path, ignore_type=False, marker="markers"):
    if not ignore_type:
        raise NotImplementedError(
            "Parsing cell types is not yet implemented for YAML files. "
            "Currently the only option is to merge them. Please try again with"
            " 'ignore_type=True'."
        )
    else:
        with open(cells_file_path, "r") as yml_file:
            data = yaml.safe_load(yml_file)
        cells = []
        for cell_type in list(data.keys()):
            type_dict = data[cell_type]
            if marker in type_dict.keys():
                for cell in type_dict[marker]:
                    cells.append(Cell(cell, Cell.UNKNOWN))
    return cells


def get_cells_dir(cells_file_path, cell_type=None):
    cells = []
    for file in os.listdir(cells_file_path):
        # ignore hidden files
        if not file.startswith("."):
            cells.append(Cell(file, cell_type))
    return cells


def save_cells(
    cells,
    xml_file_path,
    save_csv=False,
    indentation_str="  ",
    artifact_keep=True,
):
    # Assume always save xml file, and maybe save other formats
    cells_to_xml(
        cells,
        xml_file_path,
        indentation_str=indentation_str,
        artifact_keep=artifact_keep,
    )

    if save_csv:
        csv_file_path = replace_extension(xml_file_path, "csv")
        cells_to_csv(cells, csv_file_path)


def cells_to_xml(
    cells, xml_file_path, indentation_str="  ", artifact_keep=True
):
    xml_data = make_xml(cells, indentation_str, artifact_keep=artifact_keep)
    with open(xml_file_path, "w") as xml_file:
        xml_file.write(str(xml_data, "UTF-8"))


def cells_xml_to_df(xml_file_path):
    cells = get_cells(xml_file_path)
    return cells_to_dataframe(cells)


def cells_to_dataframe(cells: List[Cell]) -> pd.DataFrame:
    return pd.DataFrame([c.to_dict() for c in cells])


def cells_to_csv(cells, csv_file_path):
    df = cells_to_dataframe(cells)
    df.to_csv(csv_file_path)


def make_xml(cell_list, indentration_str, artifact_keep=True):
    root = EtElement("CellCounter_Marker_File")
    image_properties = EtElement("Image_Properties")
    file_name = EtElement("Image_Filename")
    file_name.text = "placeholder.tif"
    image_properties.append(file_name)
    root.append(image_properties)

    marker_data = EtElement("Marker_Data")
    current_type = EtElement("Current_Type")
    current_type.text = str(1)  # TODO: check
    marker_data.append(current_type)

    cell_dict = make_type_dict(cell_list)

    # if artifacts exist, do something with them (convert/delete)
    if Cell.ARTIFACT in cell_dict:
        cell_dict = deal_with_artifacts(cell_dict, artifact_keep=artifact_keep)

    for cell_type, cells in cell_dict.items():
        type_el = EtElement("Type")
        type_el.text = str(cell_type)
        mt = EtElement("Marker_Type")
        mt.append(type_el)
        for cell in cells:
            mt.append(cell.to_xml_element())
        marker_data.append(mt)
    root.append(marker_data)

    return pretty_xml(root, indentration_str)


def deal_with_artifacts(cell_dict, artifact_keep=True):
    # What do we want to do with the artifacts (if they exist)?
    # Might want to keep them for training
    if artifact_keep:
        logging.debug("Keeping artifacts")
        # Change their type
        for idx, artifact in enumerate(cell_dict[Cell.ARTIFACT]):
            cell_dict[Cell.ARTIFACT][idx].type = Cell.UNKNOWN
        # Add them to "cell_type = UNKNOWN" list
        cell_dict[Cell.UNKNOWN].extend(cell_dict[Cell.ARTIFACT])
    else:
        logging.debug("Removing artifacts")
    del cell_dict[Cell.ARTIFACT]  # outside if, needs to be run regardless
    return cell_dict


def make_type_dict(cell_list):
    types = sorted(set([cell.type for cell in cell_list]))
    return {
        cell_type: [cell for cell in cell_list if cell.type == cell_type]
        for cell_type in types
    }


def pretty_xml(elem, indentation_str="  "):
    ugly_xml = ElementTree.tostring(elem, "utf-8")
    md_parsed = minidom.parseString(ugly_xml)
    return md_parsed.toprettyxml(indent=indentation_str, encoding="UTF-8")


def find_relevant_tiffs(tiffs, cell_def):
    cells = [UntypedCell(tiff) for tiff in tiffs]
    if os.path.isdir(cell_def):
        relevant_cells = set(
            [UntypedCell(pos_from_file_name(f)) for f in os.listdir(cell_def)]
        )
    else:
        relevant_cells = set(
            [UntypedCell.from_cell(cell) for cell in get_cells(cell_def)]
        )
    return [
        tiffs[pos]
        for (pos, cell) in enumerate(cells)
        if cell in relevant_cells
    ]



from brainglobe_utils.general import list as list_tools

list_with_empty = ["test1", "test 2", " ", "", "test4", ""]

list_without_empty = ["test1", "test 2", " ", "test4"]


def test_remove_empty_string():
    assert (
        list_tools.remove_empty_string(list_with_empty) == list_without_empty
    )


def test_unique_elements_list():
    list_in = [1, 2, 2, "a", "b", 1, "a", "dog"]
    unique_list = [1, 2, "a", "b", "dog"]
    assert list_tools.unique_elements_lists(list_in) == unique_list



import pytest
from qtpy.QtWidgets import (
    QFormLayout,
    QHBoxLayout,
    QLabel,
    QPushButton,
    QVBoxLayout,
    QWidget,
)

from brainglobe_utils.qtpy.collapsible_widget import (
    CollapsibleWidget,
    CollapsibleWidgetContainer,
)

WIDGET_TITLE = "Title"


@pytest.fixture(scope="class")
def generic_widget() -> QWidget:
    widget = QWidget()
    return widget


@pytest.fixture(scope="class")
def collapsible_widget() -> CollapsibleWidget:
    collapsible_widget = CollapsibleWidget(WIDGET_TITLE)
    return collapsible_widget


@pytest.fixture(scope="class")
def collapsible_widget_container() -> CollapsibleWidgetContainer:
    collapsible_widget_container = CollapsibleWidgetContainer()
    return collapsible_widget_container


def test_collapsible_widget_empty(qtbot, collapsible_widget):
    qtbot.addWidget(collapsible_widget)

    assert collapsible_widget.text() == WIDGET_TITLE
    assert not collapsible_widget.isExpanded()
    assert collapsible_widget.content().layout().count() == 0


def test_collapsible_widget_filled(qtbot, collapsible_widget):
    label_str = "test"
    collapsible_widget.addWidget(QPushButton(label_str))

    qtbot.addWidget(collapsible_widget)

    assert collapsible_widget.content().layout().count() == 1


def test_collapsible_widget_click_once(qtbot, collapsible_widget):
    qtbot.addWidget(collapsible_widget)

    with qtbot.waitSignal(
        collapsible_widget.toggled_signal_with_self, timeout=1000
    ) as blocker:
        collapsible_widget._toggle_btn.click()

    assert blocker.args == [collapsible_widget, True]


@pytest.mark.parametrize("num_clicks", [2, 5, 100])
def test_collapsible_widget_click_multiple(
    qtbot, collapsible_widget, num_clicks
):
    """
    Test that the toggled_signal_with_self signal is emitted the correct
    number of times and with the correct parameters when clicked multiple
    times.
    """
    qtbot.addWidget(collapsible_widget)

    current_state = collapsible_widget.isExpanded()

    # Function to check the signal is emitted with the correct parameters
    # The signaller should be the collapsible widget and the state should
    # match the state tracked manually in current_state
    def check_signal_valid(signaller, state):
        return signaller == collapsible_widget and state == current_state

    with qtbot.waitSignals(
        [collapsible_widget.toggled_signal_with_self] * num_clicks,
        check_params_cbs=[check_signal_valid] * num_clicks,
        timeout=1000,
    ):
        for _ in range(num_clicks + 1):
            # Invert the current state to simulate a toggle manually
            current_state = not current_state
            collapsible_widget._toggle_btn.click()


def test_collapsible_widget_container(qtbot, collapsible_widget_container):
    qtbot.addWidget(collapsible_widget_container)

    assert collapsible_widget_container.layout().count() == 0
    assert len(collapsible_widget_container.collapsible_widgets) == 0


@pytest.mark.parametrize(
    "layout", [QVBoxLayout(), QHBoxLayout(), QFormLayout()]
)
def test_collapsible_widget_container_add_collapsible_widget(
    qtbot, collapsible_widget_container, generic_widget, layout
):
    qtbot.addWidget(collapsible_widget_container)

    generic_widget.setLayout(layout)
    generic_widget.layout().addWidget(QLabel("test"))
    generic_widget.layout().addWidget(QPushButton("test"))

    collapsible_widget_container.add_widget(
        generic_widget, collapsible=True, widget_title=WIDGET_TITLE
    )

    # Check that the widget was added and is a CollapsibleWidget
    assert collapsible_widget_container.layout().count() == 1
    assert isinstance(
        collapsible_widget_container.layout().itemAt(0).widget(),
        CollapsibleWidget,
    )
    # Check that the widget is collapsed and contains the generic widget
    assert (
        collapsible_widget_container.collapsible_widgets[0].content()
        is generic_widget
    )
    assert (
        collapsible_widget_container.layout().itemAt(0).widget().text()
        == WIDGET_TITLE
    )
    assert (
        not collapsible_widget_container.layout()
        .itemAt(0)
        .widget()
        .isExpanded()
    )
    assert len(collapsible_widget_container.collapsible_widgets) == 1


@pytest.mark.parametrize("widget_type", [QLabel, QPushButton])
def test_collapsible_widget_container_add_not_collapsible_widget(
    qtbot, collapsible_widget_container, widget_type
):
    qtbot.addWidget(collapsible_widget_container)

    collapsible_widget_container.add_widget(
        widget_type(WIDGET_TITLE), collapsible=False
    )

    assert collapsible_widget_container.layout().count() == 1
    assert isinstance(
        collapsible_widget_container.layout().itemAt(0).widget(), widget_type
    )
    assert len(collapsible_widget_container.collapsible_widgets) == 0


@pytest.mark.parametrize(
    "layout, collapsible",
    [
        (QVBoxLayout(), True),
        (QHBoxLayout(), True),
        (QFormLayout(), True),
        (QVBoxLayout(), False),
        (QHBoxLayout(), False),
        (QFormLayout(), False),
    ],
)
def test_collapsible_widget_container_add_remove_widgets(
    qtbot, collapsible_widget_container, generic_widget, layout, collapsible
):
    qtbot.addWidget(collapsible_widget_container)

    generic_widget.setLayout(layout)
    generic_widget.layout().addWidget(QLabel("test"))
    generic_widget.layout().addWidget(QPushButton("test"))

    collapsible_widget_container.add_widget(
        generic_widget, collapsible=collapsible, widget_title=WIDGET_TITLE
    )

    assert collapsible_widget_container.layout().count() == 1
    # Convert collapsible to int (False -> 0, True -> 1)
    # to check if collapsible_widgets is empty
    assert len(collapsible_widget_container.collapsible_widgets) == int(
        collapsible
    )

    collapsible_widget_container.remove_widget(generic_widget)

    assert collapsible_widget_container.layout().count() == 0
    assert len(collapsible_widget_container.collapsible_widgets) == 0


def test_collapsible_widget_container_add_remove_diff_widgets(
    qtbot, generic_widget, collapsible_widget_container
):
    qtbot.addWidget(collapsible_widget_container)
    other_widget = QLabel("test")

    collapsible_widget_container.add_widget(
        generic_widget, collapsible=True, widget_title=WIDGET_TITLE
    )
    collapsible_widget_container.add_widget(other_widget, collapsible=False)

    assert collapsible_widget_container.layout().count() == 2
    assert len(collapsible_widget_container.collapsible_widgets) == 1

    collapsible_widget_container.remove_widget(generic_widget)

    assert collapsible_widget_container.layout().count() == 1
    assert len(collapsible_widget_container.collapsible_widgets) == 0

    collapsible_widget_container.remove_widget(other_widget)

    assert collapsible_widget_container.layout().count() == 0
    assert len(collapsible_widget_container.collapsible_widgets) == 0


def test_collapsible_widget_container_remove_widget_not_found(
    qtbot, generic_widget, collapsible_widget_container
):
    with pytest.raises(ValueError):
        collapsible_widget_container.remove_widget(generic_widget)


@pytest.mark.parametrize(
    "num_collapsible_widgets, num_other_widgets, index_expanded",
    [(2, 4, 1), (5, 1, 3), (10, 0, 9)],
)
def test_collapsible_widget_container_update_drawers(
    qtbot,
    collapsible_widget_container,
    num_collapsible_widgets,
    num_other_widgets,
    index_expanded,
):
    """
    Test that the collapsible widgets in the container are expanded/collapsed
    correctly when the _update_drawers function is called.

    A container is created with num_collapsible_widgets collapsible widgets
    and num_other_widgets other widgets. The index_expanded collapsible widget
    is expanded and the state of each widget is checked. index_expanded is
    then incremented (modulo num_collapsible_widgets) to simulate toggling
    each collapsible widget in sequence. Checks the state of each widget after
    every iteration.
    """
    qtbot.addWidget(collapsible_widget_container)
    collapsible_widgets = []
    non_collapsible_widgets = []

    # Add collapsible widgets and other widgets to the container alternating
    # until the correct number of each type of widget has been added
    for i in range(num_collapsible_widgets + num_other_widgets):
        if i % 2 == 0 and len(non_collapsible_widgets) < num_other_widgets:
            collapsible_widget_container.add_widget(
                QLabel("test"), collapsible=False
            )
            non_collapsible_widgets.append(
                collapsible_widget_container.layout().itemAt(i).widget()
            )
        elif len(collapsible_widgets) < num_collapsible_widgets:
            collapsible_widget_container.add_widget(
                QLabel("test"), collapsible=True
            )
            collapsible_widgets.append(
                collapsible_widget_container.layout().itemAt(i).widget()
            )
        else:
            collapsible_widget_container.add_widget(
                QLabel("test"), collapsible=False
            )
            non_collapsible_widgets.append(
                collapsible_widget_container.layout().itemAt(i).widget()
            )

    for _ in range(num_collapsible_widgets):
        collapsible_widgets[index_expanded]._toggle_btn.click()

        for i in range(num_collapsible_widgets):
            assert collapsible_widgets[i].isExpanded() == (i == index_expanded)

        for widget in non_collapsible_widgets:
            assert not widget.isHidden()

        index_expanded = (index_expanded + 1) % num_collapsible_widgets



from brainglobe_utils.qtpy.logo import header_widget


def test_logo(qtbot):
    package_name = "test"
    package_tagline = "Tagline"
    package_tutorial = "test.html"

    header = header_widget(package_name, package_tagline, package_tutorial)

    qtbot.addWidget(header)

    expected_strings_logo = [package_name, "brainglobe.png"]
    expected_strings_docs = [
        package_tagline,
        package_tutorial,
        "https://brainglobe.info",
        "https://github.com",
    ]

    assert header.layout().count() == 2

    for logo_string in expected_strings_logo:
        assert logo_string in header.layout().itemAt(0).widget().text()

    for docs_string in expected_strings_docs:
        assert docs_string in header.layout().itemAt(1).widget().text()



from importlib.resources import files

from qtpy.QtWidgets import QGroupBox, QHBoxLayout, QLabel, QWidget


def _docs_links_widget(
    package_name: str,
    package_tagline: str,
    tutorial_file_name: str,
    parent: QWidget = None,
):
    _docs_links_html = f"""
    <h3>
    <p>{package_tagline}</p>
    <p><a href="https://brainglobe.info" style="color:gray;">Website</a></p>
    <p><a href="https://brainglobe.info/tutorials/{tutorial_file_name}" style="color:gray;">Tutorial</a></p>
    <p><a href="https://github.com/brainglobe/{package_name}" style="color:gray;">Source</a></p>
    </h3>
    """  # noqa: E501
    docs_links_widget = QLabel(_docs_links_html, parent=parent)
    docs_links_widget.setOpenExternalLinks(True)
    return docs_links_widget


def _logo_widget(package_name: str, parent: QWidget = None):
    brainglobe_logo = files("brainglobe_utils").joinpath("qtpy/brainglobe.png")

    _logo_html = f"""
    <h1>
    <img src="{brainglobe_logo}"width="100">
    <p>{package_name}</p>
    <\h1>
    """  # noqa W605

    return QLabel(_logo_html, parent=None)


def header_widget(
    package_name: str,
    package_tagline: str,
    tutorial_file_name: str,
    parent: QWidget = None,
):
    """
    Render HTML in a QGroupBox with a BrainGlobe logo and links to the docs
     and source code.

    Parameters
    ----------
    package_name : str
        The name of the package, e.g. "brainrender-napari"
    package_tagline : str
        The tagline for the package
    tutorial_file_name : str
        The name of the tutorial file (must include .html),
         e.g. "brainrender-qtpy.html": str
    parent : QWidget
        The parent widget, defaults to None

    Returns
    -------
    QGroupBox
        A QGroupBox with the rendered HTML containing the logo and links
    """

    box = QGroupBox(parent)
    box.setFlat(True)
    box.setLayout(QHBoxLayout())
    box.layout().addWidget(_logo_widget(package_name, parent=box))
    box.layout().addWidget(
        _docs_links_widget(
            package_name, package_tagline, tutorial_file_name, parent=box
        )
    )

    return box



import pytest

from brainglobe_utils.general.exceptions import CommandLineInputError


def raise_CommandLineInputError():
    raise CommandLineInputError("Error")


def test_CommandLineInputError():
    with pytest.raises(CommandLineInputError) as e:
        raise_CommandLineInputError()
    assert str(e.value) == "Error"



import glob
import logging
import os
import platform
import shutil
import subprocess
from pathlib import Path
from tempfile import gettempdir

import psutil
from natsort import natsorted
from slurmio import slurmio
from tqdm import tqdm

from brainglobe_utils.general.string import get_text_lines

# On Windows, max_workers must be less than or equal to 61
# https://docs.python.org/3/library/concurrent.futures.html#processpoolexecutor
MAX_PROCESSES_WINDOWS = 61


def replace_extension(file, new_extension, check_leading_period=True):
    """
    Replaces the file extension of a given file
    :param str file: Input file with file extension to replace
    :param str new_extension: New file extension
    :param bool check_leading_period: If True, any leading period of the
    new extension is removed, preventing "file..txt"
    :return str: File with new file extension
    """
    if check_leading_period:
        new_extension = remove_leading_character(new_extension, ".")
    return os.path.splitext(file)[0] + "." + new_extension


def remove_leading_character(string, character):
    """
    If "string" starts with "character", strip that leading character away.
    Only removes the first instance
    :param string:
    :param character:
    :return: String without the specified, leading character
    """
    if string.startswith(character):
        return string[1:]
    else:
        return string


def ensure_directory_exists(directory):
    """
    If a directory doesn't exist, make it. Works for pathlib objects, and
    strings.
    :param directory:
    """
    if isinstance(directory, str):
        if not os.path.exists(directory):
            os.makedirs(directory)
    elif isinstance(directory, Path):
        directory.mkdir(exist_ok=True)


def get_sorted_file_paths(file_path, file_extension=None, encoding=None):
    """
    Sorts file paths with numbers "naturally" (i.e. 1, 2, 10, a, b), not
    lexiographically (i.e. 1, 10, 2, a, b).
    :param str file_path: File containing file_paths in a text file,
    or as a list.
    :param str file_extension: Optional file extension (if a directory
     is passed)
    :param encoding: If opening a text file, what encoding it has.
    Default: None (platform dependent)
    :return: Sorted list of file paths
    """

    if isinstance(file_path, list):
        return natsorted(file_path)

    # assume if not a list, is a file path
    file_path = Path(file_path)
    if file_path.suffix == ".txt":
        return get_text_lines(file_path, sort=True, encoding=encoding)
    elif file_path.is_dir():
        if file_extension is None:
            file_path = glob.glob(os.path.join(file_path, "*"))
        else:
            file_path = glob.glob(
                os.path.join(file_path, "*" + file_extension)
            )
        return natsorted(file_path)

    else:
        message = (
            "Input file path is not a recognised format. Please check it "
            "is a list of file paths, a text file of these paths, or a "
            "directory containing image files."
        )
        raise NotImplementedError(message)


def check_path_in_dir(file_path, directory_path):
    """
    Check if a file path is in a directory
    :param file_path: Full path to a file
    :param directory_path: Full path to a directory the file may be in
    :return: True if the file is in the directory
    """
    directory = Path(directory_path)
    parent = Path(file_path).parent
    return parent == directory


def get_num_processes(
    min_free_cpu_cores=2,
    ram_needed_per_process=None,
    fraction_free_ram=0.1,
    n_max_processes=None,
    max_ram_usage=None,
    enforce_single_cpu=True,
):
    """
    Determine how many CPU cores to use, based on a minimum number of cpu cores
    to leave free, and an optional max number of processes.

    Cluster computing aware for the SLURM job scheduler, and not yet
    implemented for other environments.
    :param int min_free_cpu_cores: How many cpu cores to leave free
    :param float ram_needed_per_process: Memory requirements per process. Set
    this to ensure that the number of processes isn't too high.
    :param float fraction_free_ram: Fraction of the ram to ensure stays free
    regardless of the current program.
    :param int n_max_processes: Maximum number of processes
    :param float max_ram_usage: Maximum amount of RAM (in bytes)
    to use (allthough available may be lower)
    :param enforce_single_cpu: Ensure that >=1 cpu cores are chosen
    :return: Number of processes to
    """
    logging.debug("Determining the maximum number of CPU cores to use")

    n_cpu_cores = get_cores_available()
    n_cpu_cores = n_cpu_cores - min_free_cpu_cores

    logging.debug(f"Number of CPU cores available is: {n_cpu_cores}")

    if ram_needed_per_process is not None:
        n_processes = limit_cores_based_on_memory(
            n_cpu_cores,
            ram_needed_per_process,
            fraction_free_ram,
            max_ram_usage,
        )
    else:
        n_processes = n_cpu_cores

    n_max_processes = limit_cpus_windows(n_max_processes)

    if n_max_processes is not None:
        if n_max_processes < n_processes:
            logging.debug(
                f"Limiting the number of processes to {n_max_processes} based"
                f" on other considerations."
            )
        n_processes = min(n_processes, n_max_processes)

    if enforce_single_cpu:
        if n_processes < 1:
            logging.debug("Forcing number of processes to be 1")
            n_processes = 1

    logging.debug(f"Setting number of processes to: {n_processes}")
    return int(n_processes)


def limit_cpus_windows(n_max_processes):
    if platform.system() == "Windows":
        if n_max_processes is not None:
            n_max_processes = min(n_max_processes, MAX_PROCESSES_WINDOWS)
    return n_max_processes


def get_cores_available():
    try:
        os.environ["SLURM_JOB_ID"]
        n_cpu_cores = slurmio.SlurmJobParameters().allocated_cores
    except KeyError:
        n_cpu_cores = psutil.cpu_count()

    return n_cpu_cores


def limit_cores_based_on_memory(
    n_cpu_cores, ram_needed_per_process, fraction_free_ram, max_ram_usage
):
    cores_w_sufficient_ram = how_many_cores_with_sufficient_ram(
        ram_needed_per_process,
        fraction_free_ram=fraction_free_ram,
        max_ram_usage=max_ram_usage,
    )
    n_processes = min(n_cpu_cores, cores_w_sufficient_ram)
    logging.debug(
        f"Based on memory requirements, up to {cores_w_sufficient_ram} "
        f"cores could be used. Therefore setting the number of "
        f"processes to {n_processes}."
    )
    return n_processes


def how_many_cores_with_sufficient_ram(
    ram_needed_per_cpu, fraction_free_ram=0.1, max_ram_usage=None
):
    """
    Based on the amount of RAM needed per CPU core for a multiprocessing task,
    work out how many CPU cores could theoretically be used based on the
    amount of free RAM. N.B. this does not relate to how many CPU cores
    are actually available.

    :param float ram_needed_per_cpu: Memory requirements per process. Set
    this to ensure that the number of processes isn't too high.
    :param float fraction_free_ram: Fraction of the ram to ensure stays free
    regardless of the current program.
    :param float max_ram_usage: The Maximum amount of RAM (in bytes)
    to use (although available may be lower)
    :return: How many CPU cores could be theoretically used based on
    the amount of free RAM
    """

    try:
        # if in slurm environment
        os.environ["SLURM_JOB_ID"]
        # Only allocated memory (not free). Assumes that nothing else will be
        # running
        free_mem = slurmio.SlurmJobParameters().allocated_memory
    except KeyError:
        free_mem = get_free_ram()

    logging.debug(f"Free memory is: {free_mem} bytes.")

    if max_ram_usage is not None:
        free_mem = min(free_mem, max_ram_usage)
        logging.debug(
            f"Maximum memory has been set as: {max_ram_usage} "
            f"bytes, so using: {free_mem} as the maximum "
            f"available memory"
        )

    free_mem = free_mem * (1 - fraction_free_ram)
    cores_w_sufficient_ram = free_mem / ram_needed_per_cpu
    return int(cores_w_sufficient_ram // 1)


def disk_free_gb(file_path):
    """
    Return the free disk space, on a disk defined by a file path.
    :param file_path: File path on the disk to be checked
    :return: Free space in GB
    """
    if platform.system() == "Windows":
        drive, _ = os.path.splitdrive(file_path)
        total, used, free = shutil.disk_usage(drive)
        return free / 1024**3
    else:
        stats = os.statvfs(file_path)
        return (stats.f_frsize * stats.f_bavail) / 1024**3


def get_free_ram():
    """
    Returns the amount of free RAM in bytes
    :return: Available RAM in bytes
    """
    return psutil.virtual_memory().available


def safe_execute_command(cmd, log_file_path=None, error_file_path=None):
    """
    Executes a command in the terminal, making sure that the output can
    be logged even if execution fails during the call.

    From https://github.com/SainsburyWellcomeCentre/amap_python by
    Charly Rousseau (https://github.com/crousseau).

    :param cmd:
    :param log_file_path:
    :param error_file_path:
    :return:
    """
    if log_file_path is None:
        log_file_path = os.path.abspath(
            os.path.join(gettempdir(), "safe_execute_command.log")
        )
    if error_file_path is None:
        error_file_path = os.path.abspath(
            os.path.join(gettempdir(), "safe_execute_command.err")
        )

    with open(log_file_path, "w") as log_file, open(
        error_file_path, "w"
    ) as error_file:
        try:
            subprocess.check_call(
                cmd, stdout=log_file, stderr=error_file, shell=True
            )
        except subprocess.CalledProcessError:
            hline = "-" * 25
            try:
                with open(error_file_path, "r") as err_file:
                    errors = err_file.readlines()
                    errors = "".join(errors)
                with open(log_file_path, "r") as _log_file:
                    logs = _log_file.readlines()
                    logs = "".join(logs)
                raise SafeExecuteCommandError(
                    "\n{0}\nProcess failed:\n {1}"
                    "{0}\n"
                    "{2}\n"
                    "{0}\n"
                    "please read the logs at {3} and {4}\n"
                    "{0}\n"
                    "command: {5}\n"
                    "{0}".format(
                        hline,
                        errors,
                        logs,
                        log_file_path,
                        error_file_path,
                        cmd,
                    )
                )
            except IOError as err:
                raise SafeExecuteCommandError(
                    f"Process failed: please read the logs at {log_file_path} "
                    f"and {error_file_path}; command: {cmd}; err: {err}"
                )


class SafeExecuteCommandError(Exception):
    pass


def delete_directory_contents(directory, progress=False):
    """
    Removes all contents of a directory
    :param directory: Directory with files to be removed
    """
    files = os.listdir(directory)
    if progress:
        for f in tqdm(files):
            os.remove(os.path.join(directory, f))
    else:
        for f in files:
            os.remove(os.path.join(directory, f))



import random
from argparse import ArgumentTypeError
from random import randint

import pytest

from brainglobe_utils.general import numerical


def test_is_even():
    even_number = random.randrange(2, 1000, 2)
    odd_number = random.randrange(1, 1001, 2)
    with pytest.raises(NotImplementedError):
        assert numerical.is_even(0)
    assert numerical.is_even(even_number)
    assert not numerical.is_even(odd_number)


def test_check_positive_float():
    pos_val = randint(1, 1000) / 100
    neg_val = -randint(1, 1000) / 100

    assert pos_val == numerical.check_positive_float(pos_val)

    with pytest.raises(ArgumentTypeError):
        assert numerical.check_positive_float(neg_val)

    assert numerical.check_positive_float(None) is None

    with pytest.raises(ArgumentTypeError):
        assert numerical.check_positive_float(None, none_allowed=False)

    assert numerical.check_positive_float(0) == 0


def test_check_positive_int():
    pos_val = randint(1, 1000)
    neg_val = -randint(1, 1000)

    assert pos_val == numerical.check_positive_int(pos_val)

    with pytest.raises(ArgumentTypeError):
        assert numerical.check_positive_int(neg_val)

    assert numerical.check_positive_int(None) is None

    with pytest.raises(ArgumentTypeError):
        assert numerical.check_positive_int(None, none_allowed=False)

    assert numerical.check_positive_int(0) == 0
